#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <unistd.h>

#include "st7789.h"

static void send_cmd(const struct driver *drv, uint8_t cmd);
static void send_data_8bit(const struct driver *drv, uint8_t data);

static void lcd_init(const struct driver *drv);
static void lcd_set_windows(const struct driver *drv, uint16_t x_start, uint16_t y_start, uint16_t x_end, uint16_t y_end);

int st7789_open(const struct driver *drv)
{
    assert(drv != NULL);
    struct st7789 *this = (struct st7789*)(drv->dev);
    int res = drv_open(this->rst);
    if(res < 0) {
        return res;
    }
    res = drv_ioctl(this->rst, 0, 1, 1);
    if(res < 0) {
        drv_close(this->rst);
        return res;
    }
    res = drv_write(this->rst, "1", 1);
    if(res < 0) {
        drv_close(this->rst);
        return res;
    }
    res = drv_open(this->dc);
    if(res < 0) {
        drv_close(this->rst);
        return res;
    }
    res = drv_ioctl(this->dc, 0, 1, 1);
    if(res < 0) {
        drv_close(this->rst);
        drv_close(this->dc);
        return res;
    }
    res = drv_write(this->dc, "1", 1);
    if(res < 0) {
        drv_close(this->rst);
        drv_close(this->dc);
        return res;
    }
    res = drv_open(this->bl);
    if(res < 0) {
        drv_close(this->rst);
        drv_close(this->dc);
        return res;
    }
    res = drv_ioctl(this->bl, 0, 1, 1);
    if(res < 0) {
        drv_close(this->rst);
        drv_close(this->dc);
        drv_close(this->bl);
        return res;
    }
    res = drv_write(this->bl, "1", 1);
    if(res < 0) {
        drv_close(this->rst);
        drv_close(this->dc);
        drv_close(this->bl);
        return res;
    }
    res = drv_open(this->spi);
    if(res < 0) {
        drv_close(this->rst);
        drv_close(this->dc);
        drv_close(this->bl);
        return res;
    }
    // hard reset
    res = drv_write(this->rst, "0", 1);
    usleep(100 * 1000);
    res = drv_write(this->rst, "1", 1);
    usleep(100 * 1000);

    lcd_init(drv);

    return 0;
}

int st7789_close(const struct driver *drv)
{
    assert(drv != NULL);
    struct st7789 *this = (struct st7789*)(drv->dev);
    drv_close(this->rst);
    drv_close(this->dc);
    drv_close(this->bl);
    drv_close(this->spi);
    return 0;
}

int st7789_write(const struct driver *drv, const char *buffer, unsigned int len)
{
    assert(drv != NULL);
    struct st7789 *this = (struct st7789*)(drv->dev);
    uint16_t *image = (uint16_t *)buffer;
    lcd_set_windows(drv, 0, 0, this->width, this->height);
    drv_write(this->dc, "1", 1);
    for (uint16_t i = 0; i < this->height; i++) {
        drv_write(this->spi, (const char *)&image[i * this->width], this->width * 2);
    }
    return len;
}

static void send_cmd(const struct driver *drv, uint8_t cmd)
{
    assert(drv != NULL);
    struct st7789 *this = (struct st7789*)(drv->dev);
    drv_write(this->dc, "0", 1);
    drv_write(this->spi, (const char* )&cmd, 1);
}

static void send_data_8bit(const struct driver *drv, uint8_t data)
{
    assert(drv != NULL);
    struct st7789 *this = (struct st7789*)(drv->dev);
    drv_write(this->dc, "1", 1);
    drv_write(this->spi, (const char* )&data, 1);
}

static void lcd_init(const struct driver *drv)
{
    send_cmd(drv, 0x3A);
    send_data_8bit(drv, 0x05);

    send_cmd(drv, 0xB2);
    send_data_8bit(drv, 0x0C);
    send_data_8bit(drv, 0x0C);
    send_data_8bit(drv, 0x00);
    send_data_8bit(drv, 0x33);
    send_data_8bit(drv, 0x33);

    send_cmd(drv, 0xB7);  //Gate Control
    send_data_8bit(drv, 0x35);

    send_cmd(drv, 0xBB);  //VCOM Setting
    send_data_8bit(drv, 0x19);

    send_cmd(drv, 0xC0); //LCM Control
    send_data_8bit(drv, 0x2C);

    send_cmd(drv, 0xC2);  //VDV and VRH Command Enable
    send_data_8bit(drv, 0x01);
    send_cmd(drv, 0xC3);  //VRH Set
    send_data_8bit(drv, 0x12);
    send_cmd(drv, 0xC4);  //VDV Set
    send_data_8bit(drv, 0x20);

    send_cmd(drv, 0xC6);  //Frame Rate Control in Normal Mode
    send_data_8bit(drv, 0x0F);
    
    send_cmd(drv, 0xD0);  // Power Control 1
    send_data_8bit(drv, 0xA4);
    send_data_8bit(drv, 0xA1);

    send_cmd(drv, 0xE0);  //Positive Voltage Gamma Control
    send_data_8bit(drv, 0xD0);
    send_data_8bit(drv, 0x04);
    send_data_8bit(drv, 0x0D);
    send_data_8bit(drv, 0x11);
    send_data_8bit(drv, 0x13);
    send_data_8bit(drv, 0x2B);
    send_data_8bit(drv, 0x3F);
    send_data_8bit(drv, 0x54);
    send_data_8bit(drv, 0x4C);
    send_data_8bit(drv, 0x18);
    send_data_8bit(drv, 0x0D);
    send_data_8bit(drv, 0x0B);
    send_data_8bit(drv, 0x1F);
    send_data_8bit(drv, 0x23);

    send_cmd(drv, 0xE1);  //Negative Voltage Gamma Control
    send_data_8bit(drv, 0xD0);
    send_data_8bit(drv, 0x04);
    send_data_8bit(drv, 0x0C);
    send_data_8bit(drv, 0x11);
    send_data_8bit(drv, 0x13);
    send_data_8bit(drv, 0x2C);
    send_data_8bit(drv, 0x3F);
    send_data_8bit(drv, 0x44);
    send_data_8bit(drv, 0x51);
    send_data_8bit(drv, 0x2F);
    send_data_8bit(drv, 0x1F);
    send_data_8bit(drv, 0x1F);
    send_data_8bit(drv, 0x20);
    send_data_8bit(drv, 0x23);

    send_cmd(drv, 0x21);  //Display Inversion On
    send_cmd(drv, 0x11);  //Sleep Out
    send_cmd(drv, 0x29);  //Display On
}

static void lcd_set_windows(const struct driver *drv, uint16_t x_start, uint16_t y_start, uint16_t x_end, uint16_t y_end)
{
    //set the X coordinates
    send_cmd(drv, 0x2A);
    send_data_8bit(drv, (x_start >> 8) & 0xFF);
    send_data_8bit(drv, x_start & 0xFF);
    send_data_8bit(drv, ((x_end  - 1) >> 8) & 0xFF);
    send_data_8bit(drv, (x_end  - 1) & 0xFF);

    //set the Y coordinates
    send_cmd(drv, 0x2B);
    send_data_8bit(drv, (y_start >> 8) & 0xFF);
    send_data_8bit(drv, y_start & 0xFF);
    send_data_8bit(drv, ((y_end  - 1) >> 8) & 0xFF);
    send_data_8bit(drv, (y_end  - 1) & 0xFF);

    send_cmd(drv, 0X2C);
}
